// mgui.cpp
// ~~~~~~~~

// Universal headers

#include "stdafx.h"
#include "mgui.h"
#include "windows.h"
#include "commctrl.h"

// Wince/Win32 specific headers

#ifdef WINCE
	//n.b. shinitextracontrols for tooltips
	#include "richedit_ce.h"
#else
	#include "richedit.h"
	#pragma comment(linker,"\"/manifestdependency:type='win32' name='Microsoft.Windows.Common-Controls' version='6.0.0.0' processorArchitecture='*' publicKeyToken='6595b64144ccf1df' language='*'\"")
#endif

// Defines

#define CACHE_THINGS g_hInstance = hInst; g_ParentHWnd = hParent; g_ID = ID

// Global variables

HFONT hfMain = (HFONT)0;	// Caches global MessageFont as an HFONT
TCHAR fontFace[32];
LONG  fontSize = 0;			// Twips, not pixels
bool bICCE = false;			// Has InitCommonControlsEx() been called for this application?

int LogheightToTwips(int lfHeight) {
	HDC hDC = GetDC(GetDesktopWindow() );
	int PPI_Y = GetDeviceCaps(hDC, LOGPIXELSY);
	ReleaseDC( GetDesktopWindow(), hDC);
	return abs( (lfHeight*72*20) / PPI_Y);
}

void SetControlFont(HWND hWnd) {
#ifndef WINCE
	if (hfMain==(HFONT)0) {
		// Retrieve system font
		NONCLIENTMETRICS ncmSystem = {0}; //n.b. changes with NT6, so set WINVER
		ncmSystem.cbSize = sizeof(ncmSystem);
		SystemParametersInfo(SPI_GETNONCLIENTMETRICS, sizeof(ncmSystem), &ncmSystem, 0);
		hfMain = CreateFontIndirect(&ncmSystem.lfMessageFont);

		// Cache facename and height in twips
		_tcscpy_s(fontFace, 32, ncmSystem.lfMessageFont.lfFaceName);
		fontSize = LogheightToTwips(ncmSystem.lfMessageFont.lfHeight);
	}
	
	//Send new font (win32 only)
	SendMessage(hWnd, WM_SETFONT, (WPARAM)hfMain, 0); 
#endif
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MGWindow
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

// CONSTRUCTION

MGWindow::MGWindow(HINSTANCE hInst) {
	HICON hIcon = NULL;
#ifndef WINCE
	hIcon = LoadIcon(NULL, IDI_APPLICATION);
#endif
	MGWindow(hInst, hIcon, NULL, L"MGWindow", L"", CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT);
}

MGWindow::MGWindow(HINSTANCE hInst, HICON hIcon, int iMenuID, LPCTSTR lpszClassName,
					  LPCTSTR lpszWindowTitle, int x, int y, int width, int height)
{
#ifndef WINCE
	if (!bICCE) {
		INITCOMMONCONTROLSEX icc = {0};
		icc.dwSize = sizeof(INITCOMMONCONTROLSEX);
		icc.dwICC = ICC_WIN95_CLASSES | ICC_STANDARD_CLASSES | ICC_TAB_CLASSES;
		InitCommonControlsEx(&icc);
		bICCE = true;
	}
#endif

    WNDCLASS wcx; 
		wcx.style = CS_HREDRAW | CS_VREDRAW;
		wcx.lpfnWndProc = MGWindow::stProc;
		wcx.cbClsExtra = 0;
		wcx.cbWndExtra = 0;
		wcx.hInstance = hInst;
		wcx.hIcon = hIcon;
		wcx.hCursor = LoadCursor(NULL, IDC_ARROW);
		wcx.hbrBackground = (HBRUSH)(COLOR_WINDOW);
		wcx.lpszMenuName = NULL;
		wcx.lpszClassName = lpszClassName;
	RegisterClass(&wcx);

	DWORD dwFlags;
#ifdef WINCE
	dwFlags = WS_VISIBLE;
#else 
	dwFlags = WS_VISIBLE | WS_OVERLAPPEDWINDOW;
#endif
	g_hWnd = CreateWindowW(lpszClassName, lpszWindowTitle, dwFlags, x, y, width, height, NULL, NULL, hInst, (void*)this);

	bActive = g_hWnd?true:false;
	if (!bActive) return;

	SetControlFont(g_hWnd);

#ifdef WINCE
	g_hBar = CommandBar_Create(hInst, g_hWnd, IDC_MGWCOMMANDBAR);
	CommandBar_InsertMenubar(g_hBar, hInst, iMenuID, 0);
	CommandBar_AddAdornments(g_hBar, 0, 0);
	CommandBar_Show(g_hBar, true);
#else
	if (iMenuID) SetMenu(g_hWnd, LoadMenu(hInst, MAKEINTRESOURCE(iMenuID)) );
#endif

	ShowWindow(g_hWnd, true);
	UpdateWindow(g_hWnd);
}

MGWindow::~MGWindow() {
	SendMessage(g_hWnd, WM_DESTROY, 0, 0);
}

// EVENT PROCS

LRESULT CALLBACK MGWindow::stProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	MGWindow* pWnd;

#ifdef WINCE
	if (uMsg == WM_CREATE) {
#else
	if (uMsg == WM_NCCREATE) {
#endif
		SetWindowLong(hwnd, GWL_USERDATA, (long)((LPCREATESTRUCT(lParam))->lpCreateParams)); //store
	}

	pWnd = (MGWindow*) GetWindowLong(hwnd, GWL_USERDATA); //retrieve

	if (pWnd) if (pWnd->IsActive()) {
		return pWnd->Proc(hwnd, uMsg, wParam, lParam);
	}
	return DefWindowProc(hwnd, uMsg, wParam, lParam);
}

LRESULT CALLBACK MGWindow::Proc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	// overrides;
	if (uMsg == WM_DESTROY) {
#ifdef WINCE
		if (g_hBar) CommandBar_Destroy(g_hBar);
#endif
		PostQuitMessage(0);
		bActive = false;
		return 0;
	}

#ifdef WINCE
	if (uMsg == WM_SIZE && g_hBar) {
		SendMessage(g_hBar, TB_AUTOSIZE, 0, 0); //params?
		CommandBar_AlignAdornments(g_hBar); //can't test...
	}
#endif

	// user map search
	if (events.empty()) return DefWindowProc(hwnd, uMsg, wParam, lParam);

	EMAP::iterator ei = events.find(uMsg);
	if (ei != events.end()) {
		ei->second(wParam, lParam);
		return FALSE;
	} else {
		return DefWindowProc(hwnd, uMsg, wParam, lParam);
	}
}

void MGWindow::RegisterEvent(UINT uMessage, void(*fnHandler)(WPARAM,LPARAM) ) {
	events[uMessage] = fnHandler;
}

// METHODS

void MGWindow::HandleEvents() {
	MSG msg;
	if (GetMessage(&msg, NULL, 0, 0) ) {
		TranslateMessage(&msg); 
		DispatchMessage(&msg); 
	}
}

HMENU MGWindow::GetHMenu() {
#ifdef WINCE
	return CommandBar_GetMenu(g_hBar, 0);
#else
	return GetMenu(g_hWnd);
#endif
}

bool MGWindow::IsActive() { return bActive; };
HWND MGWindow::GetHwnd()  { return g_hWnd; };
int MGWindow::GetWidth()  { RECT rct; GetClientRect( g_hWnd, &rct ); return rct.right - rct.left; }

int MGWindow::GetHeight() {
	RECT rct;
	GetClientRect( g_hWnd, &rct );
#ifdef WINCE
	return (rct.bottom - rct.top) - CommandBar_Height(g_hBar);
#else
	return rct.bottom - rct.top;
#endif
}

int MGWindow::GetTop() {
#ifdef WINCE
	return CommandBar_Height(g_hBar);
#else
	return 0;
#endif
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MGControl
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void MGControl::Resize(int x, int y, int width, int height) {
	SetWindowPos(g_hWnd, 0, x, y, width, height, SWP_NOZORDER);
}

void MGControl::Resize(int width, int height) {
	SetWindowPos(g_hWnd, 0, 0, 0, width, height, SWP_NOZORDER | SWP_NOMOVE);
}

void MGControl::Focus() {
	if (g_hWnd) SetFocus(g_hWnd);
}

void MGControl::SetVisible(bool bVisible) {
	ShowWindow(g_hWnd, (bVisible? SW_SHOW : SW_HIDE) );
}

RECT MGControl::rect() {
	RECT rctOut;
	memset(&rctOut, 0, sizeof(RECT) );
	GetWindowRect(g_hWnd, &rctOut);
	return rctOut;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MGEdit
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

// CONSTRUCTION

MGEdit::MGEdit(HINSTANCE hInst, HWND hParent, int ID, int x, int y, int width, int height, bool bMultiline) {
	CACHE_THINGS;
	g_hWnd = CreateWindowEx(WS_EX_CLIENTEDGE, L"EDIT", L"",
		WS_VISIBLE | WS_CHILD | ES_AUTOVSCROLL | ES_AUTOHSCROLL | (bMultiline? ES_MULTILINE:0),
		x, y, width, height, hParent, (HMENU) ID, hInst, NULL);
	if (!g_hWnd) return;

	SetControlFont(g_hWnd);
	SetWindowLongW(g_hWnd, GWL_USERDATA, (LONG)this);
	oldProc = SetWindowLongW(g_hWnd, GWL_WNDPROC, (LONG)(&stProc));
}

// EVENT PROCS

void MGEdit::RegisterEvent(UINT uMsg, void(*fnHandler)(WPARAM,LPARAM) ) {
	events[uMsg] = fnHandler;
}

void MGEdit::Proc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	if (events.empty()) return;
	EMAP::iterator ei = events.find(uMsg);
	if (ei != events.end()) ei->second(wParam, lParam);
}

LRESULT CALLBACK MGEdit::stProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	MGEdit* mge = (MGEdit*)GetWindowLongW(hwnd, GWL_USERDATA);
	if (mge) {
		mge->Proc(hwnd, uMsg, wParam, lParam);
		return CallWindowProc((WNDPROC)(mge->oldProc), hwnd, uMsg, wParam, lParam);
	} else {
		return DefWindowProc(hwnd, uMsg, wParam, lParam); //has minimal effect
	}
}

// METHODS

void MGEdit::TextSet(TCHAR* szText) {
	SetDlgItemText(g_ParentHWnd, g_ID, szText);
}

void MGEdit::TextGet(TCHAR *outBuff, int charcount) {
	memset(outBuff, 0, charcount * sizeof(TCHAR));
	GetDlgItemText(g_ParentHWnd, g_ID, outBuff, charcount - 1);
}

int MGEdit::TextLength() {
	return GetWindowTextLength(g_hWnd);
}

void MGEdit::SelRange(int start, int end) {
	SendMessageW(g_hWnd, EM_SETSEL, start, end);
}

void MGEdit::SelEnd() {
	// this trick is necessary (and correct)
	SendMessageW(g_hWnd, EM_SETSEL, 0,  -1);
	SendMessageW(g_hWnd, EM_SETSEL, -1, -1);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MGRichEdit
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

MGRichEdit::MGRichEdit(HINSTANCE hInst, HWND hParent, int ID, int x, int y, int width, int height, bool bReadOnly) {
	CACHE_THINGS;
	if (!LoadLibrary(L"riched20.dll")) {
		MessageBox(hParent, L"Couldn't load richedit20.dll!", L"", 0);
		return;
	}

	g_hWnd = CreateWindowEx(WS_EX_STATICEDGE, L"RichEdit20W", L"", 
		ES_MULTILINE | ES_NOHIDESEL | WS_VISIBLE | WS_CHILD | WS_TABSTOP | WS_VSCROLL | (bReadOnly?ES_READONLY:0),
		x, y, width, height, hParent, (HMENU)ID, hInst, NULL);

	if (!g_hWnd) {
		DWORD err = GetLastError();
		TCHAR szError[255];
		wsprintf(szError, L"Couldn't create edit control!\nError #%d", (int)err);
		MessageBox(hParent, szError, L"", 0);
		return;
	}

	SelSetEnd();
	SelFontSet(RGB(0, 0, 0), false, false, false);
}

void MGRichEdit::Clear() {
	SendMessage(g_hWnd, WM_SETTEXT, 0, (LPARAM) L"" );
}

void MGRichEdit::SelTextGet(TCHAR* outbuff, int charcount) {
	TEXTRANGE tr; CHARRANGE cr;
	SendMessage(g_hWnd, EM_EXGETSEL, 0, (LPARAM)&cr);
	tr.chrg = cr;
	memset(outbuff, 0, charcount * sizeof(TCHAR));
	if (charcount > (cr.cpMax - cr.cpMin)) {
		tr.lpstrText = outbuff;
		SendMessage(g_hWnd, EM_GETTEXTRANGE, 0, (LPARAM)&tr);
	}
}

void MGRichEdit::SelTextSet(TCHAR* szText) {
	SendMessage(g_hWnd, EM_REPLACESEL, TRUE, (LPARAM)szText);
}

void MGRichEdit::SelSetRange(LONG lMin, LONG lMax) {
	SendMessage(g_hWnd, EM_SETSEL, (WPARAM)lMin, (LPARAM)lMax);
}

int  MGRichEdit::SelStartGet() {
	CHARRANGE cr;
	SendDlgItemMessage(g_ParentHWnd, g_ID, EM_EXGETSEL, 0, (LPARAM)&cr);
	return cr.cpMin;
}

int  MGRichEdit::SelEndGet() {
	CHARRANGE cr;
	SendDlgItemMessage(g_ParentHWnd, g_ID, EM_EXGETSEL, 0, (LPARAM)&cr);
	return cr.cpMax;
}

void MGRichEdit::SelSetEnd() {
#ifndef WINCE
	SendMessage(g_hWnd, EM_SETSEL, (WPARAM) -1, (LPARAM) -1);
#else 
	DWORD dMin=0, dMax=0;
	SendMessage(g_hWnd, WM_SETREDRAW, (WPARAM) FALSE, 0);			// block updates
	SendMessage(g_hWnd, EM_SETSEL, (WPARAM) -1,  (LPARAM) -1);		// select all
	SendMessage(g_hWnd, EM_GETSEL, (WPARAM)dMin, (LPARAM)dMax);		// get end
	SendMessage(g_hWnd, EM_SETSEL, (WPARAM)dMax-1, (LPARAM)dMax);	// set start to end
	SendMessage(g_hWnd, WM_SETREDRAW, (WPARAM) TRUE, 0);			// allow updates
	RECT rctWnd = {0}; GetWindowRect(g_hWnd, &rctWnd);
	InvalidateRect(g_hWnd, &rctWnd, FALSE);							// force update
#endif
}

void MGRichEdit::SelFontSet(DWORD rgb, bool bBold, bool bIta, bool bUL, TCHAR* szFont, long size) {
	CHARFORMAT cf = {0};
	cf.cbSize = sizeof(CHARFORMAT);
	SendDlgItemMessage(g_ParentHWnd, g_ID, EM_GETCHARFORMAT, SCF_SELECTION, (LPARAM)&cf);	
	cf.dwEffects = (bBold?CFE_BOLD:0) | (bIta?CFE_ITALIC:0) | (bUL?CFE_UNDERLINE:0);
	cf.crTextColor = (COLORREF)rgb;
#ifdef WINCE
	cf.dwMask = CFM_BOLD | CFM_ITALIC | CFM_UNDERLINE | CFM_COLOR;
#else
	cf.dwMask = CFM_BOLD | CFM_ITALIC | CFM_UNDERLINE | CFM_COLOR | CFM_FACE | CFM_SIZE;
	_tcscpy_s(cf.szFaceName, LF_FACESIZE-1, szFont);
	cf.yHeight = size;
#endif	
	SendDlgItemMessage(g_ParentHWnd, g_ID, EM_SETCHARFORMAT, SCF_SELECTION, (LPARAM)&cf);
}

void MGRichEdit::SelFontSet(DWORD rgb, bool bBold, bool bIta, bool bUL) {
	SelFontSet(rgb, bBold, bIta, bUL, fontFace, fontSize);
}

int  MGRichEdit::TextLength() {
	GETTEXTLENGTHEX gtl;
	gtl.codepage = 1200; //unicode codepage, as per msdn
	gtl.flags = GTL_CLOSE | GTL_NUMCHARS;
	return (int) SendMessage(g_hWnd, EM_GETTEXTLENGTHEX, (WPARAM) &gtl, 0);
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MGButton
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

MGButton::MGButton(HINSTANCE hInst, HWND hParent, int ID, int x, int y, int width, int height, TCHAR *szCaption) {
	CACHE_THINGS;
	g_hWnd = CreateWindowW(L"BUTTON", szCaption, 
		WS_VISIBLE | WS_CHILD | BS_TEXT | BS_VCENTER,
		x, y, width, height, hParent, (HMENU) ID, hInst, NULL);
	SetControlFont(g_hWnd);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MGListBox
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

// CONSTRUCTION

MGListBox::MGListBox(HINSTANCE hInst, HWND hParent, int ID, int x, int y, int width, int height, bool bAutoSort, bool bEvents) {
	CACHE_THINGS;
	g_hWnd = CreateWindowEx(WS_EX_STATICEDGE, L"LISTBOX", L"", 
		LBS_NOINTEGRALHEIGHT | LBS_NOTIFY | WS_CHILD | WS_VSCROLL | WS_VISIBLE | (bAutoSort?LBS_SORT:0),
		x, y, width, height, hParent, (HMENU) ID, hInst, NULL);
	if (!g_hWnd) return;

	SetControlFont(g_hWnd);
	if (bEvents) {
		SetWindowLongW(g_hWnd, GWL_USERDATA,(LONG)this);
		oldProc = SetWindowLongW(g_hWnd, GWL_WNDPROC, (LONG)stProc);
	}
}

// EVENT PROCS

LRESULT CALLBACK MGListBox::stProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	MGListBox* mglb = (MGListBox*) GetWindowLongW(hwnd, GWL_USERDATA);
	if (mglb) {
		mglb->Proc(hwnd, uMsg, wParam, lParam);
		return CallWindowProc((WNDPROC)mglb->oldProc, hwnd, uMsg, wParam, lParam);
	} else {
		return DefWindowProcW(hwnd, uMsg, wParam, lParam);
	}
}

void MGListBox::Proc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	if (events.empty()) return;
	EMAP::iterator ei = events.find(uMsg);
	if (ei != events.end()) ei->second(wParam, lParam);
}

void MGListBox::RegisterEvent(UINT uMsg, void(*fnHandler)(WPARAM,LPARAM) ) {
	events[uMsg] = fnHandler;
}

int MGListBox::SelectedIdx() {
	return (int) SendDlgItemMessageW(g_ParentHWnd, g_ID, LB_GETCURSEL, 0, 0);
}

// METHODS

void MGListBox::Clear() {
	SendDlgItemMessageW(g_ParentHWnd, g_ID, LB_RESETCONTENT, 0, 0);
}

void MGListBox::ItemAdd(TCHAR* szItem) {
	SendMessage(g_hWnd, LB_ADDSTRING, 0, (LPARAM)szItem);
}

int MGListBox::ItemFind(TCHAR* szItem) {
	return (int) SendMessage(g_hWnd, LB_FINDSTRING, -1, (LPARAM)szItem); //returns -1 on error
}

int MGListBox::ItemCount() {
	return (int) SendMessage(g_hWnd, LB_GETCOUNT, 0, 0);
}

void MGListBox::ItemRemove(int idx) {
	SendMessage(g_hWnd, LB_DELETESTRING, idx, 0);
}

void MGListBox::ItemGet(int idx, TCHAR* outbuff, int charcount) {
	int len = (int)SendDlgItemMessageW(g_ParentHWnd, g_ID, LB_GETTEXTLEN, idx, 0); //returns a charcount
	if (len != LB_ERR && len < charcount) {
		memset(outbuff, 0, charcount * sizeof(TCHAR) );
		SendDlgItemMessageW(g_ParentHWnd, g_ID, LB_GETTEXT, idx, (LPARAM)outbuff);
	}
}

RECT MGListBox::ItemGetRect(int idx) {
	RECT rctBuff = {0};
	SendMessage(g_hWnd, LB_GETITEMRECT, idx, (LPARAM)&rctBuff);
	return rctBuff;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MGToolTip
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

MGToolTip::MGToolTip(HINSTANCE hInst, HWND hParent, RECT rct, int iMaxWidth, TCHAR *szMessage) {

	//cache

	memset(this->szMessage, 0, 256 * sizeof(TCHAR) );
	_tcscpy_s(this->szMessage, 256-1, szMessage);

	//create

	g_hWnd = CreateWindowEx(WS_EX_TOPMOST, TOOLTIPS_CLASS, NULL,
		WS_POPUP | TTS_NOPREFIX | TTS_ALWAYSTIP | TTF_SUBCLASS,
		CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
		hParent, NULL, hInst, NULL);

	//topmost

	SetWindowPos(g_hWnd, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);

	//register toolinfo

	TOOLINFO ti = {0};
	ti.cbSize   = sizeof(TOOLINFO);
	ti.uFlags   = TTF_SUBCLASS;
	ti.lpszText = this->szMessage;
	ti.hinst    = hInst;
	ti.rect     = rct;
	ti.hwnd     = hParent;

	SendMessage(g_hWnd, TTM_ADDTOOL, 0, (LPARAM)&ti);
	SendMessage(g_hWnd, TTM_ACTIVATE, (WPARAM)TRUE, 0);

	// allow multiline

	if (iMaxWidth) {
		RECT rctParent = {0}; int width = 0;
		GetWindowRect(hParent, &rctParent);
		width = rctParent.right - rctParent.left;
		if (width) SendMessage(g_hWnd, TTM_SETMAXTIPWIDTH, 0, iMaxWidth);
	}
}

MGToolTip::~MGToolTip() {
	if (g_hWnd) { DestroyWindow(g_hWnd); }
}
