// mgui.h
// ~~~~~~

#include "stdafx.h"
#include <map>

#define IDC_MGWCOMMANDBAR -2							// for MGWindow-WinCE
typedef std::map<UINT,void(*)(WPARAM,LPARAM)> EMAP;		// for subclassing

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

int  LogheightToTwips(int lfHeight);
void SetControlFont(HWND hWnd);

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class MGWindow {
private:
	HWND g_hWnd;
	HWND g_hBar;
	bool bActive;
	EMAP events;

	LRESULT CALLBACK Proc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

public:
	MGWindow(HINSTANCE hInst);
	MGWindow(HINSTANCE hInst, HICON hIcon, int iMenuID, LPCTSTR lpszClassName,
	  LPCTSTR lpszWindowTitle, int x, int y, int width, int height);
	~MGWindow();

	static LRESULT CALLBACK stProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

	void RegisterEvent(UINT uMessage, void(*fnHandler)(WPARAM,LPARAM) );
	void HandleEvents();

	bool  IsActive();
	int   GetWidth();
	int   GetHeight();
	int   GetTop();
	HWND  GetHwnd();
	HMENU GetHMenu();

};

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class MGControl {
protected:
	HWND g_ParentHWnd;
	HINSTANCE g_hInstance;
	HWND g_hWnd;
	int g_ID;

public:
	void Resize(int x, int y, int width, int height);
	void Resize(int width, int height);
	void Focus();

	void SetVisible(bool bVisible);
	
	RECT rect();
	const HWND hWnd()	{ return g_hWnd; }
	const int  ID()		{ return g_ID; }
};

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class MGEdit : public MGControl {
private:
	void Proc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
	LONG oldProc;
	EMAP events;

public:
	MGEdit(HINSTANCE hInst, HWND hParent, int ID, int x, int y, int width, int height, bool bMultiline);
	
	static LRESULT CALLBACK stProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
	void RegisterEvent(UINT uMsg, void(*fnHandler)(WPARAM,LPARAM) );

	void TextSet(TCHAR* szText);
	void TextAppend(TCHAR* szText);
	void TextGet(TCHAR* outBuff, int charcount);
	int  TextLength();

	void SelRange(int start, int end);
	void SelEnd();
};


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class MGRichEdit : public MGControl {
public:
	MGRichEdit(HINSTANCE hInst, HWND hParent, int ID, int x, int y, int width, int height, bool bReadOnly);

	void Clear();

	void SelTextGet(TCHAR* outbuff, int charcount);
	void SelTextSet(TCHAR* szText);
	void SelFontSet(DWORD rgb, bool bBold, bool bIta, bool bUL);
	void SelFontSet(DWORD rgb, bool bBold, bool bIta, bool bUL, TCHAR* szFont, long size);

	void SelSetRange(LONG lMin, LONG lMax);
	void SelSetEnd();
	int  SelStartGet();
	int  SelEndGet();
	

	int  TextLength();
};

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class MGButton : public MGControl {
public:
	MGButton(HINSTANCE hInst, HWND hParent, int ID, int x, int y, int width, int height, TCHAR* szCaption);
};

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class MGListBox : public MGControl {
private:
	void Proc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
	LONG oldProc;
	EMAP events;
	
public:
	MGListBox(HINSTANCE hInst, HWND hParent, int ID, int x, int y, int width, int height, bool bAutoSort, bool bEvents);

	static LRESULT CALLBACK stProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
	void RegisterEvent(UINT uMsg, void(*fnHandler)(WPARAM,LPARAM) );
	
	int  SelectedIdx();
	void Clear();

	void ItemAdd(TCHAR* szItem);
	int  ItemCount();
	int  ItemFind(TCHAR* szItem);
	void ItemRemove(int idx);
	void ItemGet(int idx, TCHAR* outbuff, int charcount);
	RECT ItemGetRect(int idx);
};

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class MGToolTip {
private:
	HWND g_hWnd;
	TCHAR szMessage[256];
public:
	MGToolTip(HINSTANCE hInst, HWND hParent, RECT rct, int iMaxWidth, TCHAR* szMessage);
	~MGToolTip();
};

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class MGTabStrip {
};