/*
  NMDCHub (ansi)
  
*/

#include "stdafx.h"
#include "string.h"
#include "nmdc.h"


/////////////////////////////////////////////////////////
// CONSTRUCTION + DECONSTRUCTION
/////////////////////////////////////////////////////////


NMDCHub::NMDCHub(char* szAddress, int iPort, char* szNick, uint64 iShareSize) {
	Initialise(szAddress, iPort, szNick, iShareSize);
}

NMDCHub::NMDCHub(char* szAddress, char* szNick, uint64 iShareSize) {
	// Check for a port specifier in szAddress
	char* colon = strrchr(szAddress, ':');
	if (colon) {
		char szAddr[BUFS] = {0};
		char szPort[BUFS] = {0};
		strncpy_s(szAddr, BUFS-1, szAddress, (int)(colon-szAddress));
		strcpy_s(szPort, BUFS-1, colon+1);
		int iPort = atoi(szPort);
		Initialise(szAddr, iPort,  szNick, iShareSize);
	} else {
		Initialise(szAddress, 411, szNick, iShareSize);
	}
}

void NMDCHub::Initialise(char* szAddress, int iPort, char* szNick, uint64 iShareSize) {
	bUsedPassword = false;
	memset(&self, 0, sizeof(self) );
	strcpy_s(self.szNick, sizeof(self.szNick), szNick);
	self.iShareSize = iShareSize;

	sock.Initialize();
	sock.SetNonblocking();
	sock.Open((const uint8*) szAddress, iPort);
	bSockInitialised = true;
	push_event(NMDC_DEBUG_INFO, "CONN", szAddress);
}

void NMDCHub::Disconnect() {
	if (bSockInitialised) sock.Close();
}


/////////////////////////////////////////////////////////
// PUBLIC METHODS
/////////////////////////////////////////////////////////


void NMDCHub::SayPublic(char* szMessage) {
	//todo: sanitise
	char szBuff[BUFL];
	sprintf_s(szBuff, BUFL-1, "<%s> %s|", self.szNick, szMessage);	
	SayExplicit(szBuff);
}

void NMDCHub::SayPrivate(char* szUsername, char* szMessage) {
	//todo: sanitise
	char szBuff[BUFL];
	sprintf_s(szBuff, BUFL-1, "$To: %s From: %s $<%s> %s|", szUsername, self.szNick, self.szNick, szMessage);
	SayExplicit(szBuff);
}

void NMDCHub::Login(char* szPassword) {
	char szBuff[BUFS];
	sprintf_s(szBuff, BUFS-1, "$MyPass %s|", szPassword);
	SayExplicit(szBuff);
	bUsedPassword = true;
}

void NMDCHub::SayExplicit(char* szMessage) {
	if (!bSockInitialised) return;
	sock.Send((const uint8*) szMessage, strlen(szMessage));
	push_event(NMDC_DEBUG_INFO, "SEND", szMessage);
}

bool NMDCHub::Connected() {
	if (!bSockInitialised) return false;
	int sockState = sock.GetSocketError();
	return (sockState == sock.SocketEinprogress ||
		    sockState == sock.SocketSuccess ||
			sockState == sock.SocketEwouldblock);
}

int NMDCHub::GetUserCount() {
	return (int)userlist.size();
}

void NMDCHub::GetUserNicks(NMDCUser* arrayDest, int maxcount) {
	memset(arrayDest, 0, maxcount * sizeof(NMDCUser) );
	UMAP::iterator ui = userlist.begin();
	int idx = 0;
	for (;;) {
		arrayDest[idx] = ui->second;		
		if (idx++ == maxcount || ++ui == userlist.end()) break;
	}
}

bool NMDCHub::GetUserInfo(char* szUsername, NMDCUser* out) {
	UMAP::iterator ui;
	std::string ssUser(szUsername);
	ui = userlist.find(ssUser);
	if (ui != userlist.end()) {
		memcpy(out, &(ui->second), sizeof(NMDCUser) );
		return true;
	} else {
		return false;
	}
}


/////////////////////////////////////////////////////////
// EVENT HANDLING
/////////////////////////////////////////////////////////


void NMDCHub::push_event(int iType, char* data1, char* data2) {
	NMDCEvent ev;
	ev.type = iType;
	strncpy_s(ev.data1, BUFS-1, data1, BUFS-1);
	strncpy_s(ev.data2, BUFL-1, data2, BUFS-1);
	events.push(ev);
}

NMDCEvent NMDCHub::NextEvent() {
	process_messages();

	NMDCEvent next;
	memset(&next, 0, sizeof(NMDCEvent));
	if (events.empty()) {
		next.type = NMDC_NONE;
	} else {
		next = events.front();
		events.pop();
	}
	return next;
}


/////////////////////////////////////////////////////////
// PROTOCOL HANDLERS
/////////////////////////////////////////////////////////


void NMDCHub::process_messages() {
	if (!bSockInitialised) return;
	int br = sock.Receive(BUFL);
	if (br>0) {
		char* szBuff = new char[br+1];
		memset(szBuff, 0, br+1);
		memcpy(szBuff, sock.GetData(), br);

		push_event(NMDC_DEBUG_INFO, "RECV", szBuff);

		char* ctxPM = NULL;
		char* szMessage = strtok_s(szBuff, "|", &ctxPM);

		while(szMessage != 0) {
			switch (szMessage[0]) {
				case '$':
					char szCommand[BUFL];
					strcpy_s(szCommand, BUFL-1, szMessage);
					handle_command(szCommand); //give them a copy
					break;

				case '<':
					char szUserName[BUFS]; memset(szUserName, 0, BUFS); 
					char szUserSaid[BUFL]; memset(szUserSaid, 0, BUFL);
					char* idx; idx = 0; idx = strchr(szMessage, '>');
					if (idx>0) {
						strncpy_s(szUserName, BUFS - 1, szMessage + 1, (int)(idx - szMessage) - 1);
						strncpy_s(szUserSaid, BUFL - 1, idx + 2, BUFL);
						push_event(NMDC_PUBLIC_CHAT, szUserName, szUserSaid);
					}
					break;

				case '*':
				default:
					push_event(NMDC_SYSTEM_MSG, "\0", szMessage);
					break;
			}				
			
			szMessage = strtok_s(NULL, "|", &ctxPM); //next message
		}

		delete szBuff; szBuff = NULL;
	}
}

void NMDCHub::handle_command(char* szCommand) {
	int clen = strlen(szCommand);

	//tokenise command
	char* ctxHC; char* szCommandNext;
	szCommandNext = strtok_s(szCommand, " ", &ctxHC);
	szCommandNext = strtok_s(NULL, "", &ctxHC); //remaining message
	
	//switch command
	char szSay[4096] = {0};

	if (strcmp(szCommand, "$Lock")==0) {
		push_event(NMDC_SYSTEM_MSG, "\0", "Found a hub.");
		char key[BUFS];	parse_lock(szCommandNext, key, BUFS);
		sprintf_s(szSay, BUFL,
			     "$Supports NoGetINFO UserIP2|$Key %s|$ValidateNick %s|",
				 key, self.szNick);
		SayExplicit(szSay);

	} else if(strcmp(szCommand, "$Hello")==0) {
		if (strcmp(szCommandNext, self.szNick)==0) {
			//we just joined
			SayExplicit("$Version 1,0091|");
			SayExplicit("$GetNickList|");
			sprintf_s(szSay, BUFL,
				     "$MyINFO $ALL %s <ApexDC++ V:1.0.1,M:A,H:1/0/0,S:5>$ $10  $$%d$|",
					 self.szNick, self.iShareSize);
			SayExplicit(szSay);
			sprintf_s(szSay, BUFL, "Joined the hub as '%s'.", self.szNick);
			push_event(NMDC_SYSTEM_MSG, "\0", szSay);
		} else {
			//add + report the user here, and silently update on myinfo
			joinUser(szCommandNext);
		}

	} else if(strcmp(szCommand, "$HubName")==0) {
		sprintf_s(szSay, BUFL, "Now talking on '%s'.", szCommandNext);
		push_event(NMDC_SYSTEM_MSG, "\0", szSay);

	} else if(strcmp(szCommand, "$Supports")==0) { //safely ignorable
		sprintf_s(szSay, BUFL, "The server supports extensions: %s", szCommandNext);
		push_event(NMDC_DEBUG_INFO, "\0", szSay);

	} else if(strcmp(szCommand, "$ValidateDenide")==0) {
		if (bUsedPassword) {
			push_event(NMDC_SYSTEM_MSG, "\0", "Invalid password!");
		} else {
			push_event(NMDC_SYSTEM_MSG, "\0", "This nick is already in use!");
		}

	} else if(strcmp(szCommand, "$UserIP")==0) { //safely ignorable
		sprintf_s(szSay, BUFL, "The server detected our IP as %s.", szCommandNext);
		push_event(NMDC_DEBUG_INFO, "\0", szSay);

	} else if(strcmp(szCommand, "$GetPass")==0) {
		push_event(NMDC_NEED_PASS, "\0", "\0");

	} else if(strcmp(szCommand, "$MyINFO")==0) {
		joinUser(parse_myinfo(szCommandNext) );

	} else if(strcmp(szCommand, "$Quit")==0) {
		UMAP::iterator ei = userlist.find(szCommandNext);
		if (ei != userlist.end()) userlist.erase(ei); 
		push_event(NMDC_USER_PART, szCommandNext, "\0");

	} else if(strcmp(szCommand, "$UserCommand")==0) { //safely ignorable
		sprintf_s(szSay, BUFL, "A usercommand is avaliable, described as '%s'.", szCommandNext);
		push_event(NMDC_DEBUG_INFO, "\0", szSay);

	} else if(strcmp(szCommand, "$To:")==0) {
		events.push( parse_to(szCommandNext) );

	} else if(strcmp(szCommand, "$UserList")==0) {
		push_event(NMDC_DEBUG_INFO, "\0", "Got a user list.");

	} else if(strcmp(szCommand, "$NickList")==0) {
		push_event(NMDC_DEBUG_INFO, "\0", "Got a nicklist.");
		parse_nicklist(szCommandNext);		

	} else if(strcmp(szCommand, "$OpList")==0) {
		push_event(NMDC_DEBUG_INFO, "\0", "Got an operator list.");

	} else {
		sprintf_s(szSay, 4096, "The server sent an unknown command '%s'.", szCommand);
		push_event(NMDC_SYSTEM_MSG, "\0", szSay);

	}
	
}


/////////////////////////////////////////////////////////
// PROTOCOL HELPERS
/////////////////////////////////////////////////////////


void NMDCHub::joinUser(NMDCUser usr) {
	bool bUserIsNew = false;
	std::string ssNick(usr.szNick);
	if (userlist.find(ssNick) == userlist.end()) {
		bUserIsNew = true;
	}
	userlist[ssNick] = usr;
	if (bUserIsNew) {
		push_event(NMDC_USER_JOIN, usr.szNick, "\0");
	} else {
		push_event(NMDC_USER_UPDATE, usr.szNick, "\0");
	}
}

void NMDCHub::joinUser(char* szNick) {
	NMDCUser usr = {0};
	strcpy_s(usr.szNick, BUFS, szNick);
	joinUser(usr);
};

void NMDCHub::parse_nicklist(char* list) {
	// $NickList usera$$userb$$userc$$
	char* context = NULL;
	char* usr = strtok_s(list, "$$", &context);
	while (usr != NULL) {
		joinUser(usr);
		usr = strtok_s(NULL, "$$", &context);
	}
};

NMDCUser NMDCHub::parse_myinfo(char* myinfo) {
	// $MyINFO $ALL <nick> <description>$ $<connection><flag>$<e-mail>$<sharesize>$|
	NMDCUser user;
	memset(&user, 0, sizeof(NMDCUser) );
	char* context = NULL;
	char* val1 = NULL;
	char* val2 = NULL;
	val1 = strtok_s(myinfo, " ", &context); // $ALL
	val1 = strtok_s(NULL, " ", &context);   // <username>
	val2 = strtok_s(NULL, "$", &context);   // <description>
	if (val2 - val1 < BUFS) memcpy(user.szNick, val1, val2-val1);
	val1 = strtok_s(NULL, "$", &context);   // <connection><flag>
	if (val1 - val2 < BUFS) memcpy(user.szDescription, val2, val1-val2);
	val1 = strtok_s(NULL, "$", &context);   // <email>
	val1 = strtok_s(NULL, "$", &context);   // <sharesize>
	if (val1 > 0) user.iShareSize = _atoi64(val1);
	return user;
}

NMDCEvent NMDCHub::parse_to(char* to) {
	// recipient From: sender $<sender> message|
	NMDCEvent pm;
	memset(&pm, 0, sizeof(NMDCEvent) );
	pm.type = NMDC_PRIVATE_MSG;

	char* context = NULL;
	char* val1 = NULL;
	char* val2 = NULL;

	val1 = strtok_s(to,   " ", &context); //recipient
	val1 = strtok_s(NULL, " ", &context); //From:
	val1 = strtok_s(NULL, " ", &context); //sender
	val2 = strtok_s(NULL, " ", &context); //$<sender>
	if (val1 != 0 && val2 != 0) {
		if (val2 - val1 < BUFS) {
			memcpy(pm.data1, val1, val2-val1);
		} else {
			//????
		}
		val1 = strtok_s(NULL, " ", &context); //<message>
		if (val1 > 0) {
			memcpy(pm.data2, val1, strlen(val1));
		} else {
			//????
		}
	} else {
		//????
	}

	return pm;
}

void NMDCHub::parse_lock(char *lock, char* destBuff, int out_buffsize) {
    int len = strlen(lock);
    char* key = new char[len + 1];

	//lock-to-key transform
	for(int i = 1; i < len; i++) { key[i] = lock[i] ^ lock[i-1]; }
    key[0] = lock[0] ^ lock[len-1] ^ lock[len-2] ^ 5;
	for(int i = 0; i < len; i++) { key[i] = ((key[i]<<4) & 0xF0) | ((key[i]>>4) & 0x0F); }

	//sanitise
	memset(destBuff, 0, out_buffsize);
	char* newkey_p = destBuff;
	for(int i = 0; i < len; i++) {
		switch(key[i]) { 
            case 0:  case 5:  case 36:  case 96:  case 124:  case 126:
                sprintf_s(newkey_p, out_buffsize-(newkey_p-destBuff)-1, "/%%DCN%03d%%/", key[i]);
                newkey_p += 10;
                break;
            default:
                *newkey_p = key[i];
                newkey_p++;
        }
    }
    *newkey_p = '\0';

	//tidy up
	delete key;	key=NULL;
}